#include "queue.h"
//#include <stdio.h>

/*
 * priority queue implementation
 */


/* initialize a block of memory as a priority queue. "mem" points to the
 * block of memory; "num_elems" is the number of queue elements in the
 * queue, and "size_elems" is the size of each element, which must be
 * >= sizeof(PriorityQueueElement)
 *
 * returns pointer to first queue element on success, NULL on failure
 *
 * NOTE: this function just initializes the raw block of memory. It
 * does *NOT* initialize a priority queue structure: use QInit for
 * that.
 */


void *QInitMem(void *blk, unsigned num_elems, unsigned size_elems)
{
    PriorityQueueMember *member;
    void *origblk = blk, *blkPrev = blk - size_elems;
    unsigned i;

    /* sanity check: the member must be big enough to hold
       a PriorityQueueMember struct */
    if (size_elems < sizeof(*member))
        return 0L;

    if (num_elems == 0) return 0L; /* another sanity check */

    /* initialize all priorities to 0 and make each element
     * point to the next in the group
     */
    for (i = 0; i < num_elems-1; i++) {
        member = blk;
        blk += size_elems; /* arithmetic on void is a GNU C extension */
        member->next = (PriorityQueueMember *) blk;
	member->prev = (PriorityQueueMember *) blkPrev;
	blkPrev += size_elems;
        member->priority = 0;
    }

    // Now fix up the last (and first) member to point to 0
    member = blk;
    member->next = (PriorityQueueMember *) 0;
    member = origblk;
    member->prev = (PriorityQueueMember *) 0;

    /* return success */
    return origblk;
}

/**********************************************************
 * Initialize a priority queue.                           *
 * "mutex" protects the queue; "initial" is the first     *
 * element in the queue                                   *
 **********************************************************/

void QInit_func(PriorityQueue *queue, PriorityQueueMember *initial)
{
    queue->list = initial;
}

/**********************************************************
 * add a member to a queue (based on its priority)        *
 * "member" is added after any queue members which already*
 * have the same or higher priority                       *
 **********************************************************/

void QAdd(PriorityQueue *q, PriorityQueueMember *new_m, int comparison)
{
    PriorityQueueMember *lastm, *m;
    unsigned long pri = new_m->priority;


    lastm = 0;
    m = q->list;

    if (comparison < 0) {
      // Add to front of queue

    } else if (comparison > 0) {
      while (m && HigherPriority(m->priority, pri)) {
        lastm = m;
        m     = m->next;
      }

    } else {
      while (m && HigherOrEqualPriority(m->priority, pri)) {
        lastm = m;
        m     = m->next;
      }
    }

    if (!lastm) {
      if (q->list) q->list->prev = new_m;
      new_m->next = q->list;
      new_m->prev = (PriorityQueueMember *) 0;
      q->list = new_m;
    } else {
      new_m->next = m;
      new_m->prev = lastm;
      lastm->next = new_m;
      if (m) m->prev = new_m;
    }
}


/**********************************************************
 * remove a member from a queue                           *
 * returns 0 on success, -1 if member was not on queue     *
 **********************************************************/

int QRemove(PriorityQueue *q, PriorityQueueMember *rm)
{
  //  PriorityQueueMember *mem;

  //  mem = q->list;
  if (rm == q->list) {
    q->list = rm->next;
    if (q->list) q->list->prev = 0;
    rm->next = 0;
    rm->prev = 0;
    return 0;

  } else {
    if (rm->prev) rm->prev->next = rm->next;
    if (rm->next) rm->next->prev = rm->prev;
    rm->next = 0;
    rm->prev = 0;
    return 0;
  }

  /*
  mem = q->list;
  if (mem == rm) {
    q->list = rm->next;
    if (q->list) q->list->prev = 0;
    rm->next = 0;
    rm->prev = 0;
    return 0;
  }
  while (mem && mem->next != rm) {
    mem = mem->next;
  }
  if (mem) {
    mem->next = rm->next;
    if (mem->next) mem->next->prev = mem;
    rm->next = 0;
    rm->prev = 0;
    return 0;
  }
  */
  return -1;  // element not found in queue
}

/* remove the head of the queue; returns NULL if queue is empty */

PriorityQueueMember *QGetHead(PriorityQueue *queue)
{
    PriorityQueueMember *m;

    m = queue->list;
    if (m) {
        queue->list = m->next;
	if (queue->list) queue->list->prev = 0;
        m->next = 0;
	m->prev = 0;
    }
    return m;
}

/* force something into the head of a queue */
void QPutHead(PriorityQueue *queue, PriorityQueueMember *m)
{
  if (queue->list) queue->list->prev = m;
  m->next = queue->list;
  queue->list = m;
  queue->list->prev = 0;
}

/* function to determine whether a queue is empty or not */
int QIsEmpty(PriorityQueue *q)
{
    return(q->list == (PriorityQueueMember *)0);
}

